#include <libregexis024fa/colored_codeset.h>

#include <assert.h>

namespace regexis024 {
    ColoredCodesetSegment::ColoredCodesetSegment(uint32_t color, uint32_t right_code): color(color), right_code(right_code) {}

    ColoredCodesetSegmentList::ColoredCodesetSegmentList() {
        first = new ColoredCodesetSegment(0, UINT32_MAX);
    }

    void ColoredCodesetSegmentList::replace_myself(const ColoredCodesetSegmentList &other) {
        assert(other.first);
        ColoredCodesetSegment** in_cur = &first;
        ColoredCodesetSegment* in_other = other.first;
        while (in_other) {
            *in_cur = new ColoredCodesetSegment(*in_other);
            in_cur = &((**in_cur).next);
            in_other = in_other->next;
        }
    }

    ColoredCodesetSegmentList::ColoredCodesetSegmentList(const ColoredCodesetSegmentList &other) {
        replace_myself(other);
    }

    void ColoredCodesetSegmentList::free_myself() {
        ColoredCodesetSegment* cur = first;
        while (cur) {
            ColoredCodesetSegment* nxt = cur->next;
            delete cur;
            cur = nxt;
        }
    }

    ColoredCodesetSegmentList::~ColoredCodesetSegmentList() {
        free_myself();
    }

    ColoredCodesetSegmentList& ColoredCodesetSegmentList::operator=(const ColoredCodesetSegmentList &other) {
        free_myself();
        replace_myself(other);
        return *this;
    }

    ColoredCodeset::ColoredCodeset(uint64_t dummy_n): DummyN(dummy_n) {
        requests = {{}};
    }

    void ColoredCodeset::split_phase(const codeset_t &X) {

        uint32_t cA = 0;
        ColoredCodesetSegment* cur_seg = list.first;

        uint32_t pi = 0;

        auto advance_old = [&]()->void{
            cA = cur_seg->right_code + 1;
            cur_seg = cur_seg->next;
        };

        /* How to use: splits are made from left to right. After each split cur_seg
             * points to the rightest among sub-segments of cur_segment. */
        auto SPLIT = [&](uint32_t code_before_split)->void {
            assert(code_before_split < cur_seg->right_code);
            ColoredCodesetSegment* new_next = new ColoredCodesetSegment(cur_seg->color, cur_seg->right_code);
            new_next->divisor_on_left = true;
            cur_seg->right_code = code_before_split;
            new_next->next = cur_seg->next;
            cur_seg->next = new_next;
            advance_old();
        };

        while (cur_seg && pi < X.size()) {
            uint32_t cB = cur_seg->right_code;
            uint32_t L = X[pi].first, R = X[pi].second;

            if (L < cA) {
                if (R != UINT32_MAX && R + 1 < cA) {
                    pi++;
                } else if (R != UINT32_MAX && R + 1 == cA) {
                    cur_seg->divisor_on_left = true;
                    pi++;
                } else if (R < cB) {
                    SPLIT(R);
                    pi++;
                } else {
                    advance_old();
                }
            } else if (L == cA) {
                cur_seg->divisor_on_left = true;
                if (R < cB) {
                    SPLIT(R);
                    pi++;
                } else {
                    advance_old();
                }
            } else if (L <= cB) {
                SPLIT(L - 1);
                if (R < cB) {
                    SPLIT(R);
                    pi++;
                } else {
                    advance_old();
                }
            } else {
                advance_old();
            }
        }
    }

    void ColoredCodeset::apply_divisor(const codeset_t &X) {
        split_phase(X);
        size_t X_id = nxt_request_id++;
        size_t m = requests.size();
        size_t bm = m;
        std::vector<bool> skipped(bm, false);
        std::vector<bool> overlapped(bm, false);
        {
            bool inside = false;
            ColoredCodesetSegment* cur = list.first;
            while (cur) {
                inside = (inside != cur->divisor_on_left);
                if (inside) {
                    overlapped[cur->color] = true;
                } else {
                    skipped[cur->color] = true;
                }
                cur = cur->next;
            }
        }
        std::vector<uint32_t> alt_color(bm, 0);
        for (size_t i = 0; i < bm; i++) {
            if (skipped[i] && overlapped[i]) {
                alt_color[i] = m++;
                requests.push_back(requests[i]);
                if (X_id >= DummyN)
                    requests.back().push_back(X_id - DummyN);
            } else if (overlapped[i]) {
                if (X_id >= DummyN)
                    requests[i].push_back(X_id - DummyN);
            } else
                assert(skipped[i]);
        }
        {
            bool inside = false;
            ColoredCodesetSegment* cur = list.first;
            while (cur) {
                inside = (inside != cur->divisor_on_left);
                cur->divisor_on_left = false;
                uint32_t c = cur->color;
                if (inside && skipped[c] && overlapped[c]) {
                    cur->color = alt_color[c];
                }
                cur = cur->next;
            }
        }
    }

    void ColoredCodeset::get_splits_of_non_dummy(std::vector<codeset_t> &res_input,
        std::vector<std::vector<size_t>> &res_color_to_requests) {
        size_t n = requests.size();
        std::vector<ssize_t> nonclean_to_clean(n, -1);
        res_color_to_requests = {};

        for (size_t i = 0; i < n; i++) {
            if (!requests[i].empty()) {
                nonclean_to_clean[i] = res_color_to_requests.size();
                res_color_to_requests.push_back(requests[i]);
            }
        }

        ColoredCodesetSegment* cur = list.first;
        uint32_t L = 0;
        res_input.assign(res_color_to_requests.size(), {});
        while (cur) {
            size_t Sc = cur->color;
            if (nonclean_to_clean[Sc] >= 0) {
                res_input[nonclean_to_clean[Sc]].emplace_back(L, cur->right_code);
            }
            L = cur->right_code + 1;
            cur = cur->next;
        }
    }
}