#include <libregexis024fa/finite_automaton.h>
#include <assert.h>

namespace regexis024 {
    bool FA_Node::empty() {
        return type != one_char_read && type != det_char_crossroads;
    }

    void FA_Node::apply_lookahead_restriction(const codeset_t &restriction) {}

    void FA_Node::reAdd_references() {
        for (FA_Node** nxtPtr: get_all_transitions()){
            if (*nxtPtr)
                (**nxtPtr).refs++;
        }
    }

    std::vector<FA_Node **> FA_Node::get_all_transitions() {
        return {};
    }

    std::vector<FA_Node **> FA_Node::get_all_empty_valid_transitions() {
        return {};
    }

    std::vector<FA_Node **> FA_NodePathPart::get_all_transitions() {
        return {&nxt_node};
    }

    std::vector<FA_Node **> FA_NodePathPart::get_all_empty_valid_transitions() {
        if (nxt_node)
            return {&nxt_node};
        return {};
    }

    FA_NodeOfMatch::FA_NodeOfMatch() {type = match;}

    void FA_NodeOfMatch::apply_lookahead_restriction(const codeset_t &restriction) {
        ext_filter_added = true;
        pending_filter = restriction;
    }

    FA_NodeOfOneCharRead::FA_NodeOfOneCharRead(const codeset_t &filter, bool second_namespace) : filter(filter),
                                                                                                 second_ns(second_namespace) { type = one_char_read;}

    void FA_NodeOfOneCharRead::apply_lookahead_restriction(const codeset_t &restriction) {
        filter = intersect_sets(filter, restriction);
    }

    std::vector<FA_Node **> FA_NodeOfOneCharRead::get_all_empty_valid_transitions() {
        return {};
    }

    FA_NodeOfForking::FA_NodeOfForking() {type = forking;}

    std::vector<FA_Node **> FA_NodeOfForking::get_all_empty_valid_transitions() {
        std::vector<FA_Node**> res;
        for (size_t i = 0; i < nxt_options.size(); i++)
            if (nxt_options[i])
                res.push_back(&nxt_options[i]);
        return res;
    }

    std::vector<FA_Node **> FA_NodeOfForking::get_all_transitions() {
        std::vector<FA_Node**> res;
        for (size_t i = 0; i < nxt_options.size(); i++)
            res.push_back(&nxt_options[i]);
        return res;
    }

    FA_NodeOfLookOneBehind::FA_NodeOfLookOneBehind(const codeset_t &filter) : filter(filter) {type = look_one_behind;}

    FA_NodeOfLookOneAhead::FA_NodeOfLookOneAhead(const codeset_t &restriction) : restriction(restriction) {
        type = look_one_ahead;
    }

    FA_NodeOfTrackArrayMovImm::FA_NodeOfTrackArrayMovImm(opcode_t operation, uint16_t key, uint64_t immValue) :
            operation(operation), key(key), imm_value(immValue) {type = track_array_mov_imm;}
    //

    FA_NodeOfTrackArrayMovHalfinvariant::FA_NodeOfTrackArrayMovHalfinvariant(opcode_t operation, uint16_t key):
        operation(operation), key(key){type = track_array_mov_halfinvariant;}
    //

    void FA_NodeOfDetCharCrossroads::apply_lookahead_restriction(const codeset_t &restriction) {
        assert(false);
    }

    FA_NodeOfDetCharCrossroads::FA_NodeOfDetCharCrossroads(const std::vector<DFA_CrossroadPath> &crossroads)
            : crossroads(crossroads) {type = det_char_crossroads;}

    std::vector<FA_Node **> FA_NodeOfDetCharCrossroads::get_all_empty_valid_transitions() {
        return {};
    }

    std::vector<FA_Node **> FA_NodeOfDetCharCrossroads::get_all_transitions() {
        std::vector<FA_Node**> res;
        for (auto& tr: crossroads)
            res.push_back(&tr.nxt_node);
        return res;
    }

    /* If transferring ownership of node to container has failed, node is freed (which means it is ivalidated)
     * If this semi-ownership transfer succeded (no std::bad_alloc), then node is still valid to use, and at the end
     * of FA_Container lifetime it is guaranteed to be deleted
     */
    void FA_Container::registerNew(FA_Node *node) {
        try {
            node->nodeId = (int64_t)all.size();
            all.push_back(node);
        } catch (const std::exception& ba) {
            delete node;
            throw;
        }
    }

    DFA_CrossroadPath::DFA_CrossroadPath(const codeset_t &input, FA_Node *nxt_node): input(input),nxt_node(nxt_node) {}
    //

    FA_Container::~FA_Container() {
        for (FA_Node* n: all)
            delete n;
    }

#define bs(name, args, params) \
FA_NodeOf ## name *FA_Container::make ## name(args) {         \
FA_NodeOf ## name *node = new FA_NodeOf ## name(params);  \
registerNew(node);                                        \
return node;                                              \
}
#define COMMA ,

    bs(Match, , )
    bs(OneCharRead, const codeset_t& filter COMMA bool second_namespace, filter COMMA second_namespace)
    bs(Forking, , )
    bs(LookOneBehind, const codeset_t& filter, filter)
    bs(LookOneAhead, const codeset_t& filter, filter)
    bs(TrackArrayMovImm, opcode_t operation COMMA uint16_t key COMMA uint64_t immValue,
        operation COMMA key COMMA immValue)
    bs(TrackArrayMovHalfinvariant, opcode_t operation COMMA uint16_t key, operation COMMA key)
    bs(DetCharCrossroads, ,{})
}
