#include <libregexis024fa/graph_to_bytecode/filter.h>

#include <assert.h>
#include <algorithm>
#include <libregexis024fa/graph_to_bytecode/writing_commands.h>

namespace regexis024 {
    struct FilterSegment {
        ssize_t color;
        uint32_t L;
        uint32_t R;
    };

    std::vector<FilterSegment> convert_to_compSeg(const std::vector<codeset_t>& crossroad_codesets)
    {
        std::vector<FilterSegment> compSeg;
        std::vector<FilterSegment> seg;
        for (size_t i = 0; i < crossroad_codesets.size(); i++) {
            for (auto& p: crossroad_codesets[i]) {
                seg.push_back({(ssize_t)i, p.first, p.second});
            }
        }
        std::sort(seg.begin(), seg.end(),
            [](const FilterSegment& a, const FilterSegment& b)->bool{return a.L < b.L;});
        if (seg.empty()) {
            compSeg.push_back({-1, 0, UINT32_MAX});
        } else {
            if (seg[0].L > 0)
                compSeg.push_back({-1, 0, seg[0].L - 1});
            size_t N = seg.size();
            for (size_t i = 0; i + 1 < N; i++) {
                compSeg.push_back(seg[i]);
                assert(seg[i].R < seg[i + 1].L);
                if (seg[i].R + 1 < seg[i + 1].L)
                    compSeg.push_back({-1, seg[i].R + 1, seg[i + 1].L - 1});
            }
            compSeg.push_back(seg.back());
            if (seg.back().R < UINT32_MAX)
                compSeg.push_back({-1, seg[N - 1].R + 1, UINT32_MAX});
        }
        assert(!compSeg.empty());
        return compSeg;
    }

    /* Return whether the resulting bytecode relies on me placing [0]'th node at the end */
    void write_filter_exit(std::vector<uint8_t>& result, explicit_bookmarks& bookmark_manager,
        const std::vector<bookmark_id_t>& crossroad_marks,
        ssize_t color, bool at_the_end, bool& relies_on_proper_ending)
    {
        if (color < 0) {
            cmd_DIE(result);
        } else if (color != 0 || !at_the_end) {
            cmd_JUMP(result, bookmark_manager, crossroad_marks[color]);
        } else {
            relies_on_proper_ending = true;
        }
    }

    bool write_filter(std::vector<uint8_t>& result, explicit_bookmarks& bookmark_manager,
        const std::vector<codeset_t>& crossroad_codesets, const std::vector<bookmark_id_t>& crossroad_marks)
    {
        bool relies_on_proper_ending = false;

        std::vector<FilterSegment> compSeg = convert_to_compSeg(crossroad_codesets);
        size_t N = compSeg.size();
        struct RecFrame {
            size_t Li;
            size_t Ri;
            bool second_part = false;
            bookmark_id_t to_the_right_part;
        };

        std::vector<RecFrame> call_stack = {RecFrame{0, N - 1}};

        auto is_sandwich = [&](size_t Li, size_t Ri) -> bool {
            return Li + 2 == Ri && compSeg[Li].color == compSeg[Ri].color && compSeg[Li + 1].L == compSeg[Li + 1].R;
        };

        while (!call_stack.empty()) {
            RecFrame& cur_frame = call_stack.back();
            size_t Li = cur_frame.Li;
            size_t Ri = cur_frame.Ri;
            if (Li == Ri) {
                write_filter_exit(result, bookmark_manager, crossroad_marks, compSeg[Li].color,
                    Ri + 1 == N, relies_on_proper_ending);
                call_stack.pop_back();
            } else if (is_sandwich(Li, Ri)){
                ssize_t A = compSeg[Li].color;
                ssize_t B = compSeg[Li + 1].color;
                size_t midVal = compSeg[Li + 1].L;
                if (B < 0) {
                    assert(A >= 0);
                    bookmark_id_t b_to_end = bookmark_manager.new_bookmark();
                    cmd_JCEQUAL(result, bookmark_manager, midVal, b_to_end);
                    cmd_JUMP(result, bookmark_manager, crossroad_marks[A]);
                    bookmark_manager.land_bookmark(result, b_to_end);
                    cmd_DIE(result);
                } else {
                    cmd_JCEQUAL(result, bookmark_manager, midVal, crossroad_marks[B]);
                    write_filter_exit(result, bookmark_manager, crossroad_marks, A,
                        Ri + 1 == N, relies_on_proper_ending);
                }
                call_stack.pop_back();
            } else {
                size_t m = (Li + Ri) / 2;
                if (!cur_frame.second_part) {
                    cur_frame.to_the_right_part = bookmark_manager.new_bookmark();
                    cmd_JCGRTR(result, bookmark_manager, compSeg[m].R, cur_frame.to_the_right_part);
                    cur_frame.second_part = true;
                    /* cur_frame was just invalidated */
                    call_stack.push_back({Li, m});
                } else {
                    bookmark_manager.land_bookmark(result, cur_frame.to_the_right_part);
                    /* cur_frame was invalidated */
                    call_stack.pop_back();
                    call_stack.push_back({m + 1, Ri});
                }
            }
        }
        return relies_on_proper_ending;
    }
}
