#include <libregexis024sol/expr_compiler.h>
#include <libregexis024vm/vm_opcodes.h>
#include <libregexis024sol/utils.h>
#include <libregexis024fa/codeset.h>
#include <libregexis024fa/finite_automaton.h>
#include <libregexis024fa/fa_first_stage_fix.h>
#include <libregexis024fa/fa_make_deterministic.h>
#include <libregexis024fa/misc_fa_funcs.h>
#include <assert.h>
#include <memory>
#include <libregexis024sol/sol_misc_base.h>
#include <libregexis024sol/special_terminals.h>
#include <libregexis024sol/subexpr_fa_transformed.h>
#include <libregexis024sol/expr_parse_functions/epf.h>
#include <libregexis024sol/expr_parse_functions/command_recognition.h>
#include <libregexis024fa/graph_to_bytecode/fa_compiler.h>
#include <libregexis024sol/common_codesets.h>
/* Temporary debug measures */
#include <debugging_regexis024/debug_through_graphviz.h>

#define call_ERROR_CHECK do { if (ctx.error) { return NULL; } } while (0)
#define call_THROW(str) do { report(ctx, "regex: " str); return NULL; } while (0)
#define aux_ERROR_CHECK do { if (ctx.error) { return; } } while (0)
#define aux_THROW(str) do { report(ctx, "regex: " str); return; } while (0)

namespace regexis024 {
    /* ****************************** Top */

    const char* dfa_arg_aliases_condone[] = {"forgive", "condone", "okay", "optional", "nonimportant", "ifpossible", NULL};
    const char* dfa_arg_aliases_acerbic[] = {"acerbic", "angry", "pedantic", "nofork", "pure", "important", "fierce", NULL};

    void dfa_command_processing(REGEX_IS024_MeaningContext &ctx, ParsingContext& pctx, const Command& cmdBuf){
        if (pctx.dfa_cmd_activated){
            report(ctx, "repeating !dfa command");
            return;
        }
        pctx.dfa_cmd_activated = true;
        if (cmdBuf.arguments.empty())
            return;
        if (cmdBuf.arguments.size() == 1 && cmdBuf.arguments[0].arguments.empty()){
            const std::string& arg_name = cmdBuf.arguments[0].name;
            if (is_string_in_stringset(arg_name.c_str(), dfa_arg_aliases_acerbic)) {
                pctx.dfa_cmd_unforgiving = true;
                return;
            }
            if (is_string_in_stringset(arg_name.c_str(), dfa_arg_aliases_condone)) {
                pctx.dfa_cmd_nonimportant = true;
                return;
            }
        }
        report(ctx, "wrong arguments in !dfa command");
    }

    void select_command_processing(REGEX_IS024_MeaningContext& ctx, ParsingContext& pctx, const Command& cmdBuf){
        if (pctx.select_cmd_encountered)
            aux_THROW("repeating !select command");
        pctx.select_cmd_encountered = true;
        for (const CommandArgument& arg: cmdBuf.arguments){
            if (arg.is_empty)
                aux_THROW("wrong arguments in !select command");
            if (ctx.ktr.track_names.count(arg.name) != 0)
                aux_THROW("repeated names in !select command");
            int64_t namedThingId = static_cast<int64_t>(ctx.ktr.track_names.size());
            ctx.ktr.track_names.insert({arg.name, namedThingId});
            ctx.ktr.retrieval_info.emplace_back();
            ctx.ktr.retrieval_info.back().stored_in_sa = true;
            ctx.ktr.retrieval_info.back().stored_in_ca = false;
            bool mm = false, coll = false;
            for (const CommandArgument& argarg: arg.arguments){
#define mm_shenanigans if (mm) {aux_THROW("bad argument to !select command");} mm = true;
                if (argarg.name == "ca" || argarg.name == "col") {
                    if (coll)
                        aux_THROW("bad argument to !select command");
                    coll = true;
                    ctx.ktr.retrieval_info.back().stored_in_ca = true;
                } else if (argarg.name == "min") {
                    mm_shenanigans
                    ctx.ktr.retrieval_info.back().used_in_sifting = true;
                    ctx.ktr.retrieval_info.back().minimizing = true;
                } else if (argarg.name == "max"){
                    mm_shenanigans
                    ctx.ktr.retrieval_info.back().used_in_sifting = true;
                } else if (argarg.name == "ign") {
                    mm_shenanigans
                } else {
                    aux_THROW("wrong parameter for prioritized parameter in !select command");
                }
#undef mm_shenanigans
            }
            pctx.is_inside_of_these_sa_subexpressions.assign(ctx.ktr.retrieval_info.size(), false);
            /* Other info will be filled once a tracking-unit with such name will be actually found in regex */
        }
    }

    void jump_into_madness(ctx_t& ctx, ParsingContext& pctx, FA_Container &fa, int hn){
        while (true){
            int32_t pch = peep(ctx); aux_ERROR_CHECK;
            if (pch != U'!'){
                return;
            }
            size_t before_it = ctx.pos;
            Command cmd = command_expr_parse(ctx); aux_ERROR_CHECK;
            if (cmd.tilda){
                ctx.have_comment_tail = true;
                ctx.comment_tail_start = ctx.pos;
                ctx.pos = ctx.input_size;
            } else if (is_header_dfa_cmd(cmd)){
                dfa_command_processing(ctx, pctx, cmd);
            } else if (is_header_select_cmd(cmd)){
                if (hn != 1)
                    aux_THROW("!select command at the wrong place");
                select_command_processing(ctx, pctx, cmd);
            } else {
                assert(!is_header_cmd(cmd));
                ctx.pos = before_it;
                break;
            }
        }
    }

    chekushka TopLvl_ParseCall::firstTime(ctx_t &ctx, ParsingContext &pctx, FA_Container &fa) {
        result.assertDefault();
        jump_into_madness(ctx, pctx, fa, 1);
        if (ctx.have_comment_tail)
            return NULL;
        return std::make_unique<ForkLvl_ParseCall>(result);
    }

    chekushka TopLvl_ParseCall::afterReceive(ctx_t &ctx, ParsingContext &pctx, FA_Container &fa) {
        jump_into_madness(ctx, pctx, fa, 2);
        if (!isEnd(ctx))
            call_THROW("top lvl: EOF expected");
        return NULL;
    }

    /* ********************************* Bracket */

    chekushka BracketLvl_ParseCall::firstTime(REGEX_IS024_MeaningContext& ctx, ParsingContext& pctx, FA_Container& fa) {
        result.assertDefault();
        assert(readChar(ctx) == U'(');
        /* sequence lvl already took care about resolving name and configuring SubtrackingNameUsageInfo */
        if (namedSubexpressionId >= 0){
            assert(ctx.ktr.retrieval_info[namedSubexpressionId].type == tracking_var_types::range);
            if (ctx.ktr.retrieval_info[namedSubexpressionId].stored_in_sa){
                assert(namedSubexpressionId < (int64_t)pctx.is_inside_of_these_sa_subexpressions.size());
                if (pctx.is_inside_of_these_sa_subexpressions[namedSubexpressionId])
                    call_THROW("subexpression that selection array tracks is nested");
                pctx.is_inside_of_these_sa_subexpressions[namedSubexpressionId] = true;
            }
        }
        return std::make_unique<ForkLvl_ParseCall>(tmp_ret_buff);
    }

    chekushka BracketLvl_ParseCall::afterReceive(REGEX_IS024_MeaningContext& ctx, ParsingContext& pctx, FA_Container& fa) {
        if (peep(ctx) != U')')
            call_THROW("missing ')'");
        readChar(ctx);
        result = tmp_ret_buff;
        if (namedSubexpressionId >= 0) {
            SubtrackingNameUsageInfo& tai_slots = ctx.ktr.retrieval_info[namedSubexpressionId];
            if (tai_slots.stored_in_ca){
                assert(tai_slots.colarr_first >= 0 && tai_slots.colarr_first < UINT16_MAX);
                assert(tai_slots.colarr_second >= 0 && tai_slots.colarr_second < UINT16_MAX);
                result = join(subexpression_from_path(fa.makeTrackArrayMovHalfinvariant(
                        opcodes::MOV_COLARR_BTPOS, tai_slots.colarr_first)), result);
                result = join(result, subexpression_from_path(fa.makeTrackArrayMovHalfinvariant(
                        opcodes::MOV_COLARR_BTPOS, tai_slots.colarr_second)));
            }
            if (tai_slots.stored_in_sa){
                assert(tai_slots.selarr_first >= 0 && tai_slots.selarr_first < UINT16_MAX);
                assert(tai_slots.selarr_second >= 0 && tai_slots.selarr_second < UINT16_MAX);
                result = join(subexpression_from_path(fa.makeTrackArrayMovHalfinvariant(
                        opcodes::MOV_SELARR_CHPOS, tai_slots.selarr_first)), result);
                result = join(result, subexpression_from_path(fa.makeTrackArrayMovHalfinvariant(
                        opcodes::MOV_SELARR_CHPOS, tai_slots.selarr_second)));
                pctx.is_inside_of_these_sa_subexpressions[namedSubexpressionId] = false;
            }
        }
        return NULL;
    }

    /* ******************************* Fork */

    chekushka ForkLvl_ParseCall::firstTime(REGEX_IS024_MeaningContext& ctx, ParsingContext& pctx, FA_Container& fa) {
        result.assertDefault();
        options.emplace_back();  // Default one contains nothing. It will be overwritten
        return std::make_unique<Sequence_ParseCall>(options.back());
    }

    chekushka ForkLvl_ParseCall::afterReceive(REGEX_IS024_MeaningContext& ctx, ParsingContext& pctx, FA_Container& fa) {
        int32_t end_reason = peep(ctx); call_ERROR_CHECK;
        if (end_reason == U'|'){
            readChar(ctx);
            return firstTime(ctx, pctx, fa);
        }
        result = forkify(options, fa);
        return NULL;
    }

    void parseBody(REGEX_IS024_MeaningContext& ctx, FA_Container& fa, SubExprCompiled& result, ParsingContext& pctx){
        std::vector<std::shared_ptr<ParseCall>> callStack;
        callStack.push_back(std::make_unique<TopLvl_ParseCall>(result));
        bool first_time = true;
        while (!callStack.empty()){
            aux_ERROR_CHECK;
            auto nxt = first_time ? callStack.back()->firstTime(ctx, pctx, fa) : \
                                            callStack.back()->afterReceive(ctx, pctx, fa);
            if (nxt){
                callStack.push_back(std::move(nxt));
                first_time = true;
            } else {
                callStack.pop_back();
                first_time = false;
            }
        }
        /* Generating priority table (sifting program) */
        for (const SubtrackingNameUsageInfo& sni: ctx.ktr.retrieval_info) {
            if (!sni.discovered)
                aux_THROW("tracking tool named in !select is not used anywhere");
            if (sni.used_in_sifting) {
                assert(sni.selarr_first >= 0);
                assert((sni.type == tracking_var_types::range) == (sni.selarr_second != -1));
                pctx.priority_table.emplace_back(sni.minimizing, sni.selarr_first, sni.selarr_second, sni.type);
            }
        }
    }

    REGEX_IS024_MeaningContext::REGEX_IS024_MeaningContext(size_t inputSize, const char *input) : input_size(inputSize),
                                                                input(input) {
        CommonCodesets codeset_collection;
        FA_Container fa;
        FA_Container fa_1f;
        FA_Container fa_2f;
        SubExprCompiled result;
        ParsingContext pctx(codeset_collection);
        parseBody(*this, fa, result, pctx);
        /* CLion gone crazy here. It thinks error is always false (It doesn't know about such thing as macros) */
        if (error)
            return;

        FA_NodeOfMatch* matcher = fa.makeMatch();
        if (!result.start){
            fa.start = matcher;
        } else {
            fa.start = result.start;
            for (FA_Node** ending: result.ends)
                reattach_fa_node_edge(ending, matcher);
        }
        fa.start->refs++;

        // show_fa_with_sxiv_after_dot(fa, ktr, pctx.priority_table);  // todo debug

        REGEX_IS024_FA_FirstStageFixInfo info1 = first_stage_fix_fa(fa, fa_1f);

        // show_fa_with_sxiv_after_dot(fa_1f, ktr, pctx.priority_table);  // todo debug

        if (pctx.dfa_cmd_activated) {
            int det_err;
            int had_to_fork;
            try_determinize_fa(fa_1f, pctx.priority_table, free_selarr_tai, info1, fa_2f, det_err, had_to_fork);
            if (det_err < 0 && !pctx.dfa_cmd_nonimportant) {
                report(*this, "Unable to determinize dfa");
                return;
            }
            if (pctx.dfa_cmd_unforgiving && had_to_fork < 0) {
                report(*this, "Attempt to determinize dfa was not good enough");
                return;
            }
        } else {
            regular_second_stage_fix(fa_1f, fa_2f, info1);
        }

        // show_fa_with_sxiv_after_dot(fa_2f, ktr, pctx.priority_table);  // todo debug

        int compilation_error;
        compile_fa_to_regexis024_bytecode(compiled_program, fa_2f, pctx.priority_table, free_selarr_tai, info1, compilation_error);
        if (compilation_error) {
            report(*this, "Failed to compile graph representation to bytecode representation");
            return;
        }
    }
}
